/**
 * Fichier : PC_Epd_154b_low_memory
 * Auteur : Pascal Chour - www.pascalchour.fr
 * Date : 23 février 2018
 * (c) Pascal Chour 
 * Basé sur epd1in54b
 * Utilisable librement. Merci de préciser la source.
 * 
 *-------------------------------------------------------------------------- 
 * Cette bibliothèque permet de gérer un afficheur e-paper 1.54 pouce type B 
 * (pas de rafraichissement partiel) sur un microcontrôleur ne disposant que de
 * peu de RAM (typiquement, 2ko comme pour certains Arduino
 *
 * La bibliothèqe permet : 
 * - d'initialiser l'afficheur (cf. Waveshare)
 * - d'effacer l'écran
 * - de remplir l'écran
 * - d'afficher des lignes horizontale 
 * - d'afficher des textes
 * Le tout en utilisant très peu de RAM (moins de 300 octets de mémoire temporaire).
 *--------------------------------------------------------------------------
 * This library can managed a 1.54 inch e-paper display type B (no partial refresh) with
 * a low memory microcontroller (typically : some Arduino with 2kb RAM).
 * The functionalities of the library are :
 * - display management (init..., see Waveshare)
 * - clear the display
 * - fill the display
 * - draw horizontal line
 * - write text
 * The RAM used is lower than 300 bytes (temporary RAM).
 *--------------------------------------------------------------------------
 *
 * Some functions came from
 *  @filename   :   epd1in54b.cpp
 *  @brief      :   Implements for e-paper library
 *  @author     :   Yehui from Waveshare
 *
 *  Copyright (C) Waveshare     August 10 2017
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documnetation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to  whom the Software is
 * furished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS OR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#include "PC_Epd_154_low_memory.h"
#include <spi.h>
#include <avr/pgmspace.h>

#include "Font20.c"
#include "Font2line.c"

unsigned int reset_pin;
unsigned int dc_pin;
unsigned int cs_pin;
unsigned int busy_pin;

unsigned int CurrentByteBlack = 0;	// Position courante sur banc BLACK
unsigned int CurrentByteRed = 0;	// Position courante sur banc RED
unsigned char CurrentBank = BLACK; 	// banc courant


sFONT Font20 = {
  Font20_Table,
  17, // Width
  20, // Height
  0x19, // Base
};

sFONT Font2line = {
  Font2line_Table,
  17, // Width
  2,  // Height
  0x20, // Base, premier caractère = Espace, 2ème caractère = 0x21
};

const unsigned char lut_vcom0[] =
{
    0x0E, 0x14, 0x01, 0x0A, 0x06, 0x04, 0x0A, 0x0A,
    0x0F, 0x03, 0x03, 0x0C, 0x06, 0x0A, 0x00
};

const unsigned char lut_w[] =
{
    0x0E, 0x14, 0x01, 0x0A, 0x46, 0x04, 0x8A, 0x4A,
    0x0F, 0x83, 0x43, 0x0C, 0x86, 0x0A, 0x04
};

const unsigned char lut_b[] =
{
    0x0E, 0x14, 0x01, 0x8A, 0x06, 0x04, 0x8A, 0x4A,
    0x0F, 0x83, 0x43, 0x0C, 0x06, 0x4A, 0x04
};

const unsigned char lut_g1[] =
{
    0x8E, 0x94, 0x01, 0x8A, 0x06, 0x04, 0x8A, 0x4A,
    0x0F, 0x83, 0x43, 0x0C, 0x06, 0x0A, 0x04
};

const unsigned char lut_g2[] =
{
    0x8E, 0x94, 0x01, 0x8A, 0x06, 0x04, 0x8A, 0x4A,
    0x0F, 0x83, 0x43, 0x0C, 0x06, 0x0A, 0x04
};

const unsigned char lut_vcom1[] =
{
    0x03, 0x1D, 0x01, 0x01, 0x08, 0x23, 0x37, 0x37,
    0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

const unsigned char lut_red0[] =
{
    0x83, 0x5D, 0x01, 0x81, 0x48, 0x23, 0x77, 0x77,
    0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

const unsigned char lut_red1[] =
{
    0x03, 0x1D, 0x01, 0x01, 0x08, 0x23, 0x37, 0x37,
    0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

void DelayMs(unsigned int delaytime) {
    delay(delaytime);
}

void SpiTransfer(unsigned char data) {
    digitalWrite(cs_pin, LOW);
    SPI.transfer(data);
    digitalWrite(cs_pin, HIGH);
}

int IfInit(int cs, int dc, int rst, int busy) {
	reset_pin = rst;
    dc_pin = dc;
    cs_pin = cs;
    busy_pin = busy;
    pinMode(cs_pin, OUTPUT);
    pinMode(reset_pin, OUTPUT);
    pinMode(dc_pin, OUTPUT);
    pinMode(busy_pin, INPUT);
    SPI.beginTransaction(SPISettings(2000000, MSBFIRST, SPI_MODE0));
    SPI.begin();
    return 0;
}


int EpdInit(int cs, int dc, int rst, int busy) {
    /* this calls the peripheral hardware interface, see epdif */
    if (IfInit(cs, dc, rst, busy) != 0) {
        return -1;
    }
    /* EPD hardware init start */
    EpdReset();
    EpdSendCommand(POWER_SETTING);
    EpdSendData(0x07);
    EpdSendData(0x00);
    EpdSendData(0x08);
    EpdSendData(0x00);
    EpdSendCommand(BOOSTER_SOFT_START);
    EpdSendData(0x07);
    EpdSendData(0x07);
    EpdSendData(0x07);
    EpdSendCommand(POWER_ON);

    EpdWaitUntilIdle();

    EpdSendCommand(PANEL_SETTING);
    EpdSendData(0xcf);
    EpdSendCommand(VCOM_AND_DATA_INTERVAL_SETTING);
    EpdSendData(0x17);
    EpdSendCommand(PLL_CONTROL);
    EpdSendData(0x39);
    EpdSendCommand(TCON_RESOLUTION);
    EpdSendData(0xC8);
    EpdSendData(0x00);
    EpdSendData(0xC8);
    EpdSendCommand(VCM_DC_SETTING_REGISTER);
    EpdSendData(0x0E);

    EpdSetLutBw();
    EpdSetLutRed();
	CurrentByteBlack = 0;
	CurrentByteRed = 0;
	CurrentBank = BLACK;
    /* EPD hardware init end */

    return 0;
}

/**
 *  @brief: basic function for sending commands
 */
void EpdSendCommand(unsigned char command) {
    digitalWrite(dc_pin, LOW);
    SpiTransfer(command);
}

/**
 *  @brief: basic function for sending data
 */
void EpdSendData(unsigned char data) {
    digitalWrite(dc_pin, HIGH);
    SpiTransfer(data);
}

/**
 *  @brief: Wait until the busy_pin goes HIGH
 */
void EpdWaitUntilIdle(void) {
    while(digitalRead(busy_pin) == 0) {      //0: busy, 1: idle
        DelayMs(100);
    }
}

/**
 *  @brief: module reset.
 *          often used to awaken the module in deep sleep,
 *          see EpdSleep();
 */
void EpdReset(void) {
    digitalWrite(reset_pin, LOW);                //module reset
    DelayMs(200);
    digitalWrite(reset_pin, HIGH);
    DelayMs(200);
}

/**
 *  @brief: set the look-up tables
 */
void EpdSetLutBw(void) {
    unsigned int count;
    EpdSendCommand(0x20);         //g vcom
    for(count = 0; count < 15; count++) {
        EpdSendData(lut_vcom0[count]);
    }
    EpdSendCommand(0x21);        //g ww --
    for(count = 0; count < 15; count++) {
        EpdSendData(lut_w[count]);
    }
    EpdSendCommand(0x22);         //g bw r
    for(count = 0; count < 15; count++) {
        EpdSendData(lut_b[count]);
    }
    EpdSendCommand(0x23);         //g wb w
    for(count = 0; count < 15; count++) {
        EpdSendData(lut_g1[count]);
    }
    EpdSendCommand(0x24);         //g bb b
    for(count = 0; count < 15; count++) {
        EpdSendData(lut_g2[count]);
    }
}

void EpdSetLutRed(void) {
    unsigned int count;
    EpdSendCommand(0x25);
    for(count = 0; count < 15; count++) {
        EpdSendData(lut_vcom1[count]);
    }
    EpdSendCommand(0x26);
    for(count = 0; count < 15; count++) {
        EpdSendData(lut_red0[count]);
    }
    EpdSendCommand(0x27);
    for(count = 0; count < 15; count++) {
        EpdSendData(lut_red1[count]);
    }
}
/*
 * démarre une session d'affichage : initialise les pointeurs mémoire à 0
 * Rend True si pas d'erreur
 * Rend faux si on essaye de redémarrer du banc noir alors qu'on est sur le banc rouge 
 * (il faut faire un EpdSessionClose => Refresh d'abord)
*/
bool EpdSessionStart(int Bank) {
	CurrentByteBlack = 0;
	CurrentByteRed = 0;
	DelayMs(2);
	if (Bank == RED) {
		CurrentBank = RED;
		EpdSendCommand(DATA_START_TRANSMISSION_2);
	} else {
		if (CurrentBank == RED) {return false;}
		CurrentBank = BLACK;
		EpdSendCommand(DATA_START_TRANSMISSION_1);
	}
	DelayMs(2);
	return true;
}
/*
* Rempli l'écran d'une couleur à partir de la position courante
* Color = UNCOLORED (blanc) ou COLORED (noir ou rouge).
*/
void EpdSessionFill(bool Color) {
unsigned char temp;
	EpdSessionGoToLineRel(EPD_HEIGHT,Color);
}


/*
* Fin de session d'affichage. Refresh...
*/
void EpdSessionClose() {
	DelayMs(2);
	EpdSendCommand(DISPLAY_REFRESH);
	CurrentBank = BLACK;
	CurrentByteBlack = 0;
	CurrentByteRed = 0;
	EpdWaitUntilIdle();
}


/*
* Met l'effaceeur à blanc sur le banc BLACK et RED
*/

void EpdSessionClearDisplay() {
	EpdSessionStart(BLACK);
	EpdSessionFill(UNCOLORED);
	EpdSessionStart(RED);
	EpdSessionFill(UNCOLORED);
	EpdSessionClose();
}

/*
* positionne le pointeur en début de la ligne L (en démarrant à 0 = 1ère ligne). 
* ATTENTION : pas de retour en arrière possible. si L est inférieure ou égale
* à la ligne courante, la fonction ne fait pas d'action
* idem si L > taille écran
* La fonction rend true si elle a pu réaliser l'action demandée
* 
*/
bool EpdSessionGoToLineAbs(unsigned int L, bool Color)
{
	unsigned char temp;
	unsigned int F;
	unsigned int *CurrentByte;	
	unsigned int i,n;
	if (L >= EPD_HEIGHT) {return false;}
	if (CurrentBank == RED) {
		F = 1;
		CurrentByte = &CurrentByteRed;
	} else {
		F = 2;
		CurrentByte = &CurrentByteBlack;
	}
	L = EPD_LINE_BYTE * F * L;
	if (L < *CurrentByte) {return false;}
	n = L - *CurrentByte;
	if (Color) {temp = 0x00;} else {temp = 0xFF;}
	for (i = 0; i < n; i++) {
		EpdSendData(temp);
		*CurrentByte += 1;
	}
	return true;
}

/*
* positionne le pointeur L lignes après la ligne courante (en démarrant à 0 = 1ère ligne). 
* ATTENTION : pas de retour en arrière possible. si L est inférieure ou égale
* à la ligne courante, la fonction ne fait pas d'action
* idem si L > taille écran
* La fonction rend true si elle a pu réaliser l'action demandée
* Note : si L=0, la fonction termine la ligne courante
* Color = True pour Black ou Red = COLORED, false pour White = UNCOLORED
* 
*/
bool EpdSessionGoToLineRel(unsigned int L, bool Color)
{
	unsigned char temp;
	unsigned int F;
	unsigned int *CurrentByte;	
	unsigned int i,n;
	if (CurrentBank == RED) {
		F = 1;
		CurrentByte = &CurrentByteRed;
	} else {
		F = 2;
		CurrentByte = &CurrentByteBlack;
	}
	n = ((L+1)*EPD_LINE_BYTE*F) - (*CurrentByte % (EPD_LINE_BYTE*F));
	if (n > EPD_MEM_BYTE*F) {return false;}
	if (Color) {temp = 0x00;} else {temp = 0xFF;}
	for (i = 0; i < n; i++) {
		EpdSendData(temp);
		*CurrentByte += 1;
	}
	return true;
}

/**
 *  Ecriture de la string text.
 * Par défaut, écriture caractères Noir sur blanc (banc noir et blanc) ou rouge sur blanc (banc rouge)
 * InvertedChar = true : inversion couleur fond et couleur caractères
 * InvertedFill = true : si banc noir et blanc, remplisage noir. Si banc rouge, remplissage rouge. Sinon, remplissage blanc
 * nrFont = n° de fonte
 * nrSpace = nombre de caractères "espace" avant l'écriture du texte
 */
void EpdSessionWriteString(unsigned int nrSpace, const char * text, int nrFont, bool InvertedChar, bool InvertedFill) {
    int i, j, k, l, n;
	unsigned int char_offset;
	unsigned char buffer[EPD_WIDTH]; 
	unsigned int ibuffer;
	unsigned int itext;
	unsigned int ispace;
	unsigned char temp;
	unsigned int max;
	unsigned char mask;
	unsigned char maskDefault;
	unsigned char shift;
	unsigned int *CurrentByte;
	sFONT *CurrentFont;					// Fonte courante
	

	switch (nrFont)
	{
		case 20: CurrentFont = &Font20;	break;
		case 2:  CurrentFont = &Font2line;	break;
		default: CurrentFont = &Font20;	break;
	}
	if (CurrentBank == RED) {
		CurrentByte = &CurrentByteRed;
		maskDefault = 0x80;
		shift = 1;
	} else {
		CurrentByte = &CurrentByteBlack;
		maskDefault = 0xC0;
		shift = 2;
	}
	for (k = 0; k < CurrentFont->Height; k++) { // Pour toute la hauteur des caractères
		if (CurrentBank == RED) {
			max = CurrentByteRed+EPD_LINE_BYTE-(CurrentByteRed % (EPD_LINE_BYTE));		// 1 bit par pixel sur fond rouge
		} else {
			max = CurrentByteBlack+2*EPD_LINE_BYTE-(CurrentByteBlack % (2*EPD_LINE_BYTE)); 	// 2 bits par pixel sur fond noir et blanc
		}
		ibuffer = 0;
		itext = 0;
		ispace = nrSpace;
		while ((text[itext] != 0) || (ispace)) {  // Pour tous les caractères
			if (ispace) {
				char_offset = (' ' - CurrentFont->Base) * CurrentFont->Height * (CurrentFont->Width / 8 + (CurrentFont->Width % 8 ? 1 : 0));
			} else {
				char_offset = (text[itext] - CurrentFont->Base) * CurrentFont->Height * (CurrentFont->Width / 8 + (CurrentFont->Width % 8 ? 1 : 0));
			}
			l = CurrentFont->Width;
			for (i = 0; i <= ((CurrentFont->Width-1)/8); i++) {
				temp = pgm_read_byte(&CurrentFont->table[char_offset+i+(k*(1+((CurrentFont->Width-1) / 8)))]);
				n = 0;
				while ((n < 8) && (l > 0)) {
					buffer[ibuffer] = temp & 0x80;
					temp = (temp << 1);
					ibuffer = ibuffer+1;
					if (ibuffer >= EPD_WIDTH) {break;}
					n = n+1;
					l = l-1;
				}
			}
			if (ispace) {ispace = ispace-1;} else {itext +=1;}
		}
		
		temp = 0xFF;
		mask = maskDefault;
		for (i = 0; i < ibuffer;i++) {
			if (buffer[i]) {temp &= ~mask;}
			mask = (mask >> shift);
			if (!mask) {
				if (*CurrentByte < max) {
					if (InvertedChar) {temp = ~temp;}
					EpdSendData(temp);
					*CurrentByte += 1;
				}
				temp = 0xFF;
				mask = maskDefault;
			}
		}
		if (mask != maskDefault) {
			if (*CurrentByte < max) {
				if (InvertedChar) {temp = ~temp;}
				EpdSendData(temp);
				*CurrentByte += 1;
			}
		}
		EpdSessionGoToLineRel(0,InvertedFill); 
	}
}

void EpdDisplayFrame(const unsigned char* frame_buffer_black, const unsigned char* frame_buffer_red) {
    unsigned char temp;
    if (frame_buffer_black != NULL) {
        EpdSendCommand(DATA_START_TRANSMISSION_1);
        DelayMs(2);
        for (int i = 0; i < EPD_WIDTH * EPD_HEIGHT / 8; i++) {
            temp = 0x00;
            for (int bit = 0; bit < 4; bit++) {
                if ((pgm_read_byte(&frame_buffer_black[i]) & (0x80 >> bit)) != 0) {
                    temp |= 0xC0 >> (bit * 2);
                }
            }
            EpdSendData(temp);
            temp = 0x00;
            for (int bit = 4; bit < 8; bit++) {
                if ((pgm_read_byte(&frame_buffer_black[i]) & (0x80 >> bit)) != 0) {
                    temp |= 0xC0 >> ((bit - 4) * 2);
                }
            }
            EpdSendData(temp);
        }
        DelayMs(2);
    }
    if (frame_buffer_red != NULL) {
        EpdSendCommand(DATA_START_TRANSMISSION_2);
        DelayMs(2);
        for (int i = 0; i < EPD_WIDTH * EPD_HEIGHT / 8; i++) {
            EpdSendData(pgm_read_byte(&frame_buffer_red[i]));
        }
        DelayMs(2);
    }
    EpdSendCommand(DISPLAY_REFRESH);
    EpdWaitUntilIdle();
}

/**
 *  @brief: After this command is transmitted, the chip would enter the
 *          deep-sleep mode to save power.
 *          The deep sleep mode would return to standby by hardware reset.
 *          The only one parameter is a check code, the command would be
 *          executed if check code = 0xA5.
 *          You can use EpdInit() to awaken
 */
void EpdSleep() {
    EpdSendCommand(VCOM_AND_DATA_INTERVAL_SETTING);
    EpdSendData(0x17);
    EpdSendCommand(VCM_DC_SETTING_REGISTER);         //to solve Vcom drop
    EpdSendData(0x00);
    EpdSendCommand(POWER_SETTING);         //power setting
    EpdSendData(0x02);        //gate switch to external
    EpdSendData(0x00);
    EpdSendData(0x00);
    EpdSendData(0x00);
    EpdWaitUntilIdle();
    EpdSendCommand(POWER_OFF);         //power off
}
/*
* Rend la largeur (Width) et la hauteur (Height) en pixels de la fonte nrFont
* Rend 0 si la fonte n'existe pas
*/
int  EpdSessionGetFontWidth(int nrFont)
{
	switch (nrFont)
	{
		case 20: return Font20.Width;
		case 2:  return Font2line.Width;
		default: return 0;
	}
}
	
int  EpdSessionGetFontHeight(int nrFont)
{
	switch (nrFont)
	{
		case 20: return Font20.Height;
		case 2:  return Font2line.Height;
		default: return 0;
	}
}	



/* END OF FILE */


